# Standard library
import re

# Packages
from bs4 import BeautifulSoup
from urllib.parse import urlparse

# Regex that matches Discourse topic URL
# It is used to pull out the slug and topic_id
TOPIC_URL_MATCH = re.compile(
    r"(?:/t)?(?:/(?P<slug>[^/]+))?/(?P<topic_id>\d+)(?:/\d+)?"
)


class BaseParser(object):
    """
    Parsers used commonly by Docs and Engage pages
    """

    def _get_section(self, soup, title_text):
        """
        Given some HTML soup and the text of a title within it,
        get the content between that title and the next title
        of the same level, and return it as another soup object.

        E.g. if `soup` contains is:

        <p>Pre</p>
        <h2>My heading</h2>
        <p>Content</p>
        <h2>Next heading</h2>

        and `title_text` is "My heading", then it will return:

        <p>Content</p>
        """

        heading = soup.find(re.compile("^h[1-6]$"), text=title_text)

        if not heading:
            return None

        heading_tag = heading.name

        section_html = "".join(map(str, heading.fetchNextSiblings()))
        section_soup = BeautifulSoup(section_html, features="html.parser")

        # If there's another heading of the same level
        # get the content before it
        next_heading = section_soup.find(heading_tag)
        if next_heading:
            section_elements = next_heading.fetchPreviousSiblings()
            section_elements.reverse()
            section_html = "".join(map(str, section_elements))
            section_soup = BeautifulSoup(section_html, features="html.parser")

        return section_soup

    def _get_preamble(self, soup, break_on_title):
        """
        Given a BeautifulSoup HTML document,
        separate out the HTML at the start, up to
        the heading defined in `break_on_title`,
        and return it as a BeautifulSoup object
        """

        heading = soup.find(re.compile("^h[1-6]$"), text=break_on_title)

        if not heading:
            return soup

        preamble_elements = heading.fetchPreviousSiblings()
        preamble_elements.reverse()
        preamble_html = "".join(map(str, preamble_elements))

        return BeautifulSoup(preamble_html, features="html.parser")

    def _parse_metadata(self, index_soup, section_name):
        """
        Given the HTML soup of an index topic
        extract the metadata from the name designated
        by section_name

        This section_name section should contain a table
        (extra markup around this table doesn't matter)
        e.g.:

        <h1>Metadata</h1>
        <details>
            <summary>Mapping table</summary>
            <table>
            <tr><th>Column 1</th><th>Column 2</th></tr>
            <tr>
                <td>data 1</td>
                <td>data 2</td>
            </tr>
            <tr>
                <td>data 3</td>
                <td>data 4</td>
            </tr>
            </table>
        </details>

        This will typically be generated in Discourse from Markdown similar to
        the following:

        # Redirects

        [details=Mapping table]
        | Column 1| Column 2|
        | -- | -- |
        | data 1 | data 2 |
        | data 3 | data 4 |

        The function will return a list of dictionaries of this format:
        [
            {"column-1": "data 1", "column-2": "data 2"},
            {"column-1": "data 3", "column-2": "data 4"},
        ]
        """
        metadata_soup = self._get_section(index_soup, section_name)

        topics_metadata = []
        if metadata_soup:
            titles = [
                title_soup.text.lower().replace(" ", "_").replace("-", "_")
                for title_soup in metadata_soup.select("th")
            ]
            for row in metadata_soup.select("tr:has(td)"):
                row_dict = {}
                for index, value in enumerate(row.select("td")):
                    if value.find("a"):
                        row_dict["topic_name"] = value.find("a").text

                    # Beautiful soup renders URLs as anchors
                    # Avoid that default behaviour
                    if value.find("a") and (
                        value.find("a")["href"] == value.find("a").text
                    ):
                        value.contents[0] = value.find("a").text

                    row_dict[titles[index]] = "".join(
                        str(content) for content in value.contents
                    )

                topics_metadata.append(row_dict)

        return topics_metadata

    def _parse_url_map(
        self, index_soup, url_prefix, index_topic_id, url_section_name
    ):
        """
        Given the HTML soup of an index topic
        extract the URL mappings from a "URLs" section.

        This section could be called whatever is
        passed in `url_section_name` but it must
        contain a table of
        "Topic" to "Path" mappings
        (extra markup around this table doesn't matter)
        e.g.:

        <h1>URLs</h1>
        <details>
            <summary>Mapping table</summary>
            <table>
            <tr><th>Topic</th><th>Path</th></tr>
            <tr>
                <td><a href="https://forum.example.com/t/page/10">Page</a></td>
                <td>/cool-page</td>
            </tr>
            <tr>
                <td>
                    <a href="https://forum.example.com/t/place/11">Place</a>
                </td>
                <td>/cool-place</td>
            </tr>
            </table>
        </details>

        This will typically be generated in Discourse from Markdown similar to
        the following:

        # URLs

        [details=Mapping table]
        | Topic | Path |
        | -- | -- |
        | https://forum.example.com/t/place/11| /cool-page |
        | https://forum.example.com/t/place/11  | /cool-place |

        """

        url_soup = self._get_section(index_soup, url_section_name)
        url_map = {}
        warnings = []

        if url_soup:
            for row in url_soup.select("tr:has(td)"):
                topic_a = row.select_one("td:first-child a[href]")
                path_td = row.select_one("td:nth-child(2)")

                if not topic_a or not path_td:
                    warnings.append("Could not parse URL map item {item}")
                    continue

                topic_url = topic_a.attrs.get("href", "")
                topic_path = urlparse(topic_url).path
                topic_match = TOPIC_URL_MATCH.match(topic_path)

                pretty_path = path_td.text
                if not pretty_path.startswith("/"):
                    pretty_path = "/" + pretty_path
                if not pretty_path.startswith(url_prefix):
                    pretty_path = url_prefix + pretty_path

                if not topic_match or not pretty_path.startswith(url_prefix):
                    warnings.append("Could not parse URL map item {item}")
                    continue

                topic_id = int(topic_match.groupdict()["topic_id"])

                url_map[pretty_path] = topic_id

        # Add the reverse mappings as well, for efficiency
        ids_to_paths = dict([reversed(pair) for pair in url_map.items()])
        url_map.update(ids_to_paths)

        # Add the homepage path
        home_path = url_prefix
        if home_path != "/" and home_path.endswith("/"):
            home_path = home_path.rstrip("/")
        url_map[home_path] = index_topic_id
        url_map[index_topic_id] = home_path

        return url_map, warnings



# Packages
import flask
import os
import requests

# Local
from canonicalwebteam.discourse import DiscourseAPI, EngageParser
from vcr_unittest import VCRTestCase


class TestDiscourseAPI(VCRTestCase):
    def _get_vcr_kwargs(self):
        """
        This removes the authorization header
        from VCR so we don't record auth parameters
        """
        return {"filter_headers": ["Authorization"]}

    def setUp(self):
        """
        Set up Flask app with Discourse extension for testing
        And set up mocking for discourse.example.com
        """
        this_dir = os.path.dirname(os.path.realpath(__file__))
        template_folder = f"{this_dir}/fixtures/templates/engage.html"
        app = flask.Flask("main", template_folder=template_folder)
        self.discourse_api = DiscourseAPI(
            base_url="https://discourse.ubuntu.com/",
            session=requests.Session(),
            api_key="secretkey",
            api_username="canonical",
        )
        self.parser = EngageParser(
            api=self.discourse_api,
            index_topic_id=17229,
            url_prefix="/engage",
        )

        app = flask.Flask("main", template_folder=template_folder)
        self.client = app.test_client()
        return super(TestDiscourseAPI, self).setUp()

    def test_get_topic(self):
        """
        Check API retrieves a protected topic 17275
        """
        response = self.discourse_api.get_topic(17275)
        # # Check for success
        self.assertEqual(response["id"], 17275)



import httpretty
import json


def register_uris():
    """
    Mocks for a fake Discourse API set of endpoints
    """

    # Index page with navigation, URL map and redirects
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/34.json",
        body=json.dumps(
            {
                "id": 34,
                "category_id": 2,
                "title": "An index page",
                "slug": "an-index-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 3434,
                            "cooked": (
                                "<p>Some homepage content</p>"
                                "<h1>Navigation</h1>"
                                "<ul>"
                                '<li><a href="/t/page-a/10">Page A</a></li>'
                                '<li><a href="/t/b-page/12">B page</a></li>'
                                "</ul>"
                                "<h1>URLs</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-a/10">Page A</a></td>'
                                "<td>/a</td>"
                                "</tr><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-z/26">Page Z</a></td>'
                                "<td>/page-z</td>"
                                "</tr></tbody></table>"
                                "</div></details>"
                                "<h1>Redirects</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody>"
                                "<tr><td>/redir-a</td><td>/a</td></tr>"
                                "<tr>"
                                "  <td>/example/page</td>"
                                "  <td>https://example.com/page</td>"
                                "</tr>"
                                "</tr></tbody></table>"
                                "</div></details>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Index page with navigation only
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/35.json",
        body=json.dumps(
            {
                "id": 35,
                "category_id": 2,
                "title": "An index page",
                "slug": "an-index-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 3435,
                            "cooked": (
                                "<p>Some homepage content</p>"
                                "<h1>Navigation</h1>"
                                "<ul>"
                                '<li><a href="/t/page-a/10">Page A</a></li>'
                                '<li><a href="/t/b-page/12">B page</a></li>'
                                "</ul>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Index page with broken, clashing url map and redirects
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/36.json",
        body=json.dumps(
            {
                "id": 36,
                "category_id": 2,
                "title": "An index page",
                "slug": "an-index-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 3436,
                            "cooked": (
                                "<p>Some homepage content</p>"
                                "<h1>Navigation</h1>"
                                "<ul>"
                                '<li><a href="/t/page-a/10">Page A</a></li>'
                                '<li><a href="/t/b-page/12">B page</a></li>'
                                "</ul>"
                                "<h1>URLs</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody><tr>"
                                '<td><a href="https://discourse.example.com/t'
                                '/page-a/10">Page A</a></td>'
                                "<td>/a</td>"
                                "</tr><tr>"
                                '<td><a href="https://discourse.example.com/t'
                                '/page-z/26">Page Z</a></td>'
                                "<td>/page-z</td>"
                                "</tr></tbody></table>"
                                "</div></details>"
                                "<h1>Redirects</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Path</th><th>Location</th></tr></thead>"
                                "<tbody>"
                                "<tr>"
                                "  <td>/a</td><td>/clashing-redirect</td>"
                                "</tr>"
                                "<tr>"
                                "  <td>invalid-path</td><td>/somewhere</td>"
                                "</tr>"
                                "<tr>"
                                "  <td>/invalid-location</td>"
                                "  <td>some-domain.com/fish</td>"
                                "</tr>"
                                "<tr>"
                                "  <td>/valid</td>"
                                "  <td>/target</td>"
                                "</tr>"
                                "</tbody></table>"
                                "</div></details>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/37.json",
        body=json.dumps(
            {
                "id": 37,
                "category_id": 2,
                "title": "An index page",
                "slug": "an-index-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 3437,
                            "cooked": (
                                "<p>Some homepage content</p>"
                                "<h1>Navigation</h1>"
                                "<ul>"
                                '<li><a href="/t/page-a/50">Page A</a></li>'
                                "</ul>"
                                "<h1>URLs</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-a/50">Page A</a></td>'
                                "<td>/a</td>"
                                "</tr></tbody></table>"
                                "</div></details>"
                                "<h1>Redirects</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody>"
                                "<tr><td>/redir-a</td><td>/a</td></tr>"
                                "<tr>"
                                "  <td>/example/page</td>"
                                "  <td>https://example.com/page</td>"
                                "</tr>"
                                "</tr></tbody></table>"
                                "</div></details>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Basic topic page with minimal content
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/10.json",
        body=json.dumps(
            {
                "id": 10,
                "category_id": 2,
                "title": "Page A",
                "slug": "page-a",
                "post_stream": {
                    "posts": [
                        {
                            "id": 56,
                            "cooked": ("<p>Content of this page</p>"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Basic topic page with minimal content
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/11.json",
        body=json.dumps(
            {
                "id": 11,
                "category_id": 2,
                "title": "Page A",
                "slug": "page-a",
                "post_stream": {
                    "posts": [
                        {
                            "id": 56,
                            "cooked": ("<p>Content of this page</p>"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Basic topic page with minimal content
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/12.json",
        body=json.dumps(
            {
                "id": 12,
                "category_id": 2,
                "title": "Page A",
                "slug": "page-a",
                "post_stream": {
                    "posts": [
                        {
                            "id": 56,
                            "cooked": ("<p>Content of this page</p>"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )
    # Basic topic page with minimal content
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/42.json",
        body=json.dumps(
            {
                "id": 42,
                "category_id": 2,
                "title": "A page",
                "slug": "a-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 56,
                            "cooked": ("<p>Content of Page A</p>"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Topic with "note to editors"
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/47.json",
        body=json.dumps(
            {
                "id": 47,
                "category_id": 2,
                "title": "Note to editors",
                "slug": "note-to-editors",
                "post_stream": {
                    "posts": [
                        {
                            "id": 59,
                            "cooked": (
                                "<p>Above</p>"
                                '<aside class="quote no-group">'
                                "<blockquote>"
                                "<p>"
                                '<img title=":construction:" class="emoji">'
                                "<strong>NOTE TO EDITORS</strong>"
                                '<img title=":construction:" class="emoji">'
                                "</p>"
                                "<p> Note content </p>"
                                "<p> Note content line 2 </p>"
                                "</blockquote>"
                                "</aside>"
                                "<p>Below</p>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    # Topic with emoji-based notification
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/49.json",
        body=json.dumps(
            {
                "id": 49,
                "category_id": 2,
                "title": "Notifications",
                "slug": "notifications",
                "post_stream": {
                    "posts": [
                        {
                            "id": 64,
                            "cooked": (
                                "<p>Before</p>"
                                "<blockquote>"
                                "<p>ⓘ A notification</p>"
                                "</blockquote>"
                                ""
                                "<blockquote>"
                                "<p>"
                                '<img src="https://forum.snapcraft.io/'
                                'images/emoji/emoji_one/warning.png?v=5" '
                                'title=":warning:" class="emoji" '
                                'alt=":warning:"> '
                                "A warning"
                                "</p>"
                                "</blockquote>"
                                ""
                                "<p>After</p>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/50.json",
        body=json.dumps(
            {
                "id": 50,
                "category_id": 3,
                "title": "B page",
                "slug": "b-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 40,
                            "cooked": ("<p>Content of Page B</p>"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/99.json",
        status=404,
        content_type="application/json",
    )

    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/c/2.json",
        body=json.dumps(
            {
                "topic_list": {
                    "per_page": 3,
                    "topics": [
                        {
                            "id": 3434,
                            "slug": "super-fancy",
                            "fancy_title": "Super fancy",
                        }
                    ],
                }
            }
        ),
    )

    # Mock with URL map that has field with prefix and some with not
    httpretty.register_uri(
        httpretty.GET,
        "https://discourse.example.com/t/38.json",
        body=json.dumps(
            {
                "id": 38,
                "category_id": 2,
                "title": "An index page",
                "slug": "an-index-page",
                "post_stream": {
                    "posts": [
                        {
                            "id": 3434,
                            "cooked": (
                                "<p>Some homepage content</p>"
                                "<h1>Navigation</h1>"
                                "<ul>"
                                '<li><a href="/t/page-a/10">Page A</a></li>'
                                '<li><a href="/t/b-page/12">B page</a></li>'
                                "</ul>"
                                "<h1>URLs</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-a/10">Page A</a></td>'
                                "<td>/docs/a</td>"
                                "</tr><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-b/11">Page B</a></td>'
                                "<td>b</td>"
                                "</tr><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-b/12">Page C</a></td>'
                                "<td>c</td>"
                                "</tr><tr>"
                                '<td><a href="https://discourse.example.com/t/'
                                'page-z/26">Page Z</a></td>'
                                "<td>/page-z</td>"
                                "</tr></tbody></table>"
                                "</div></details>"
                                "<h1>Redirects</h1>"
                                '<details open="">'
                                "<summary>Mapping table</summary>"
                                '<div class="md-table">'
                                "<table>"
                                "<thead><tr>"
                                "<th>Topic</th><th>Path</th></tr></thead>"
                                "<tbody>"
                                "<tr><td>/redir-a</td><td>/a</td></tr>"
                                "<tr>"
                                "  <td>/example/page</td>"
                                "  <td>https://example.com/page</td>"
                                "</tr>"
                                "</tr></tbody></table>"
                                "</div></details>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ]
                },
            }
        ),
        content_type="application/json",
    )



import unittest
import httpretty
import requests

from canonicalwebteam.discourse.models import DiscourseAPI
from tests.fixtures.forum_mock import register_uris


class TestDiscourseAPI(unittest.TestCase):
    def setUp(self):
        httpretty.enable()
        register_uris()

        self.api = DiscourseAPI(
            base_url="https://discourse.example.com",
            session=requests.Session(),
        )

    def test_get_topic(self):
        """
        Check the DiscourseAPI object can get a topic by its ID
        """

        topic = self.api.get_topic(34)

        self.assertEqual(topic["id"], 34)
        self.assertEqual(topic["title"], "An index page")


if __name__ == "__main__":
    unittest.main()
